@@ -3,43 +3,48 @@ function get_table as well as a set of predefined CIM queries.
 """
 
 import re
+from abc import ABC, abstractmethod
 from datetime import datetime
-from typing import Dict, Iterable, List, Optional, Tuple, TypeVar, Union
+from functools import cached_property
+from typing import Dict, Iterable, List, Optional, Tuple, Union
 
 import numpy as np
 import pandas as pd
 
-from cimsparql import queries
+from cimsparql import line_queries, queries
 from cimsparql import query_support as sup
 from cimsparql import ssh_queries, sv_queries, tp_queries
-from cimsparql.constants import generating_types, impedance_variables, mrid_variable, ratings
+from cimsparql.constants import union_split
+from cimsparql.enums import (
+    ConverterTypes,
+    GeneratorTypes,
+    Impedance,
+    LoadTypes,
+    Power,
+    Rates,
+    SyncVars,
+    TapChangerObjects,
+)
 from cimsparql.type_mapper import TypeMapper
+from cimsparql.typehints import Region
 from cimsparql.url import Prefix
 
-CimModelType = TypeVar("CimModelType", bound="CimModel")
 
+class Model(Prefix, ABC):
+    def __init__(self, mapper: Optional[TypeMapper], *args, **kwargs) -> None:
+        super().__init__(*args, **kwargs)
+        self._mapper = mapper
 
-class CimModel(Prefix):
-    """Used to query with sparql queries (typically CIM)"""
+    @abstractmethod
+    def _get_table(self, *args, **kwargs) -> Tuple[pd.DataFrame, Dict[str, str]]:
+        pass
 
-    def __init__(self, mapper: TypeMapper, *args, **kwargs) -> None:
-        self._setup_client(*args, **kwargs)
-        self.mapper = mapper
+    @abstractmethod
+    def _col_map(self) -> Dict[str, str]:
+        pass
 
-    @property
-    def date_version(self) -> datetime:
-        """Activation date for this repository"""
-        try:
-            date_version = self._date_version
-        except AttributeError:
-            repository_date = self._get_table_and_convert(queries.version_date())
-            date_version = repository_date["activationDate"].values[0]
-            if isinstance(date_version, np.datetime64):
-                date_version = self._date_version = date_version.astype("<M8[s]").astype(datetime)
-        return date_version
-
-    @property
-    def mapper(self) -> TypeMapper:
+    @cached_property
+    def mapper(self) -> Optional[TypeMapper]:
         """Mapper used to convert str → type described by ontology in Graphdb
 
         Getter:
@@ -47,30 +52,153 @@ class CimModel(Prefix):
         Setter:
            Sets mapper for self. Query GraphDB if not provided (None)
         """
+        if self._mapper is None and "rdfs" in self.prefixes:
+            return TypeMapper(self)
         return self._mapper
 
-    @mapper.setter
-    def mapper(self, mapper: Optional[TypeMapper] = None) -> None:
-        if mapper is None and "rdfs" in self.prefixes:
-            self._mapper = TypeMapper(self)
-        else:
-            self._mapper = mapper
-
     def _query_with_header(self, query: str, limit: Optional[int] = None) -> str:
         query = "\n".join([self.header_str(query), query])
         if limit is not None:
             query += f" limit {limit}"
         return query
 
+    @property
+    def empty(self) -> bool:
+        """Identify empty GraphDB repo"""
+        try:
+            self._get_table("SELECT * \n WHERE { ?s ?p ?o }", limit=1)
+            return False
+        except IndexError:
+            return True
+
+    @staticmethod
+    def _assign_column_types(
+        result: pd.DataFrame, columns: Dict[str, Union[bool, str, float, int]]
+    ) -> None:
+        for column, column_type in columns.items():
+            if column_type is str:
+                continue
+            if column_type is bool:
+                result[column] = result[column].str.contains("true", flags=re.IGNORECASE)
+            else:
+                result.loc[result[column] == "None", column] = ""
+                result[column] = pd.to_numeric(result[column], errors="coerce").astype(column_type)
+
+    @classmethod
+    def col_map(cls, data_row, columns: Dict[str, str]) -> Dict[str, str]:
+        columns = columns or {}
+        col_map = cls._col_map(data_row)
+        col_map.update(columns)
+        return col_map
+
+    def get_table(
+        self,
+        query: str,
+        index: Optional[str] = None,
+        limit: Optional[int] = None,
+        map_data_types: bool = True,
+        columns: Optional[Dict] = None,
+    ) -> pd.DataFrame:
+        """Gets given table from the configured database.
+
+        Args:
+           query: to sparql server
+           index: column name to use as index
+           limit: limit number of resulting rows
+           map_data_types: gets datatypes from the configured graphdb & maps the types in the result
+                  to correct python types
+           custom_maps: dictionary of 'sparql_datatype': function to apply on columns with that
+                  type. Overwrites sparql map for the types specified.
+           columns: dictionary of 'column_name': function,
+                  uses pandas astype on the column, or applies function.
+                  Sparql map overwrites columns when available
+
+        Example:
+           >>> from cimsparql.graphdb import GraphDBClient
+           >>> from cimsparql.url import service
+           >>> gdbc = GraphDBClient(service('LATEST'))
+           >>> query = 'select * where { ?subject ?predicate ?object }'
+           >>> gdbc.get_table(query, limit=10)
+        """
+        try:
+            result, data_row = self._get_table(query, limit)
+        except IndexError:
+            return pd.DataFrame([])
+
+        if map_data_types and self.mapper is not None:
+            col_map = self.col_map(data_row, columns)
+            result = self.mapper.map_data_types(result, col_map)
+
+        if index and not result.empty:
+            result.set_index(index, inplace=True)
+        return result
+
+    @property
+    def map_data_types(self) -> bool:
+        return "cim" in self.prefixes and self.mapper.have_cim_version(self.prefixes["cim"])
+
+    @classmethod
+    def _manual_convert_types(
+        cls, df: pd.DataFrame, columns: Optional[Dict], index: Optional[str]
+    ) -> pd.DataFrame:
+        if columns is None:
+            columns = {}
+        reset_index = index is not None
+        if reset_index:
+            df.reset_index(inplace=True)
+        df = df.astype(str)
+        cls._assign_column_types(df, columns)
+        if reset_index:
+            df.set_index(index, inplace=True)
+        return df
+
+    def _get_table_and_convert(
+        self,
+        query: str,
+        limit: Optional[int] = None,
+        index: Optional[str] = None,
+        columns: Optional[Dict[str, str]] = None,
+    ) -> pd.DataFrame:
+        result = self.get_table(query, index, limit, map_data_types=True, columns=columns)
+
+        if not self.map_data_types and len(result) > 0:
+            result = self._manual_convert_types(result, columns, index)
+
+        return result
+
+
+class CimModel(Model):
+    """Used to query with sparql queries (typically CIM)"""
+
+    def __init__(self, *args, **kwargs) -> None:
+        super().__init__(*args, **kwargs)
+        self._date_version = None
+
+    @property
+    def date_version(self) -> datetime:
+        """Activation date for this repository"""
+        if self._date_version:
+            return self._date_version
+        repository_date = self._get_table_and_convert(queries.version_date())
+        self._date_version = repository_date["activationDate"].values[0]
+        if isinstance(self._date_version, np.datetime64):
+            self._date_version = self._date_version.astype("<M8[s]").astype(datetime)
+        return self._date_version
+
+    def full_model(self, dry_run: bool = False) -> Union[pd.DataFrame, str]:
+        query = queries.full_model()
+        if dry_run:
+            return self._query_with_header(query)
+        return self._get_table_and_convert(query)
+
     def bus_data(
         self,
-        region: Optional[Union[str, List[str]]] = None,
+        region: Region = None,
         sub_region: bool = False,
         limit: Optional[int] = None,
         dry_run: bool = False,
-        mrid: str = mrid_variable,
-        name: str = "?name",
-    ) -> pd.DataFrame:
+        with_dummy_buses: bool = False,
+    ) -> Union[pd.DataFrame, str]:
         """Query name of topological nodes (TP query).
 
         Args:
@@ -79,45 +207,52 @@ class CimModel(Prefix):
            limit: return first 'limit' number of rows
            dry_run: return string with sql query
         """
-        query = queries.bus_data(region, sub_region, mrid, name)
+        query = queries.bus_data(region, sub_region)
+        if with_dummy_buses:
+            dummy_bus_query = queries.three_winding_dummy_bus(region, sub_region)
+            combined = sup.combine_statements(query, dummy_bus_query, split=union_split)
+            query = sup.combine_statements(
+                sup.select_statement(["?mrid", "?name", "?busname", "?un", "?area", "?bidzone"]),
+                f"where {{{{{combined}}}}}",
+            )
         if dry_run:
             return self._query_with_header(query, limit)
-        return self._get_table_and_convert(query, limit, index=mrid[1:])
+        return self._get_table_and_convert(query, limit, index="mrid")
 
     def phase_tap_changers(
         self,
-        region: Optional[Union[str, List[str]]] = None,
+        region: Region = None,
         sub_region: bool = False,
         with_tap_changer_values: bool = True,
-        impedance: Iterable[str] = impedance_variables,
-        tap_changer_objects: Iterable[str] = ("high", "low", "neutral"),
-        mrid: str = mrid_variable,
+        impedance: Iterable[Impedance] = tuple(Impedance),
+        tap_changer_objects: Iterable[TapChangerObjects] = tuple(TapChangerObjects),
         limit: Optional[int] = None,
         dry_run: bool = False,
-    ) -> pd.DataFrame:
+    ) -> Union[pd.DataFrame, str]:
         """Get list of phase tap changers"""
         query = queries.phase_tap_changer_query(
-            region, sub_region, with_tap_changer_values, impedance, tap_changer_objects, mrid
+            region, sub_region, with_tap_changer_values, impedance, tap_changer_objects
         )
         if dry_run:
             return self._query_with_header(query, limit)
-        return self._get_table_and_convert(query, limit, index=mrid[1:])
+        return self._get_table_and_convert(query, limit, index="mrid")
 
     def loads(
         self,
-        load_type: List[str],
-        load_vars: Optional[Tuple[str]] = None,
-        region: Union[str, List[str]] = None,
+        load_type: LoadTypes,
+        load_vars: Optional[Iterable[Power]] = None,
+        region: Region = None,
         sub_region: bool = False,
-        connectivity: Optional[str] = None,
+        connectivity: Optional[Optional[str]] = None,
+        nodes: Optional[Optional[str]] = None,
         station_group_optional: bool = True,
         station_group: bool = False,
         with_sequence_number: bool = False,
         network_analysis: bool = True,
-        mrid: str = mrid_variable,
         limit: Optional[int] = None,
         dry_run: bool = False,
-    ) -> pd.DataFrame:
+        ssh_graph: Optional[str] = None,
+    ) -> Union[pd.DataFrame, str]:
         """Query load data
 
         Args:
@@ -148,27 +283,25 @@ class CimModel(Prefix):
             region,
             sub_region,
             connectivity,
+            nodes,
             station_group_optional,
             with_sequence_number,
             network_analysis,
             station_group,
             self.cim_version,
-            mrid,
+            ssh_graph,
         )
         if dry_run:
             return self._query_with_header(query, limit)
 
-        columns = {} if load_vars is None else {var: float for var in load_vars}
-        return self._get_table_and_convert(query, limit, index=mrid[1:], columns=columns)
+        return self._get_table_and_convert(query, limit, index="mrid")
 
     def wind_generating_units(
         self,
         limit: Optional[int] = None,
         network_analysis: bool = True,
-        mrid: str = mrid_variable,
-        name: str = "?name",
         dry_run: bool = False,
-    ) -> pd.DataFrame:
+    ) -> Union[pd.DataFrame, str]:
         """Query wind generating units
 
         Args:
@@ -187,28 +320,26 @@ class CimModel(Prefix):
            >>> gdbc.wind_generating_units(limit=10)
 
         """
-        query = queries.wind_generating_unit_query(network_analysis, mrid, name)
+        query = queries.wind_generating_unit_query(network_analysis)
         if dry_run:
             return self._query_with_header(query, limit)
-        float_list = ["maxP", "allocationMax", "allocationWeight", "minP"]
-        columns = {var: float for var in float_list}
-        return self._get_table_and_convert(query, limit, index=mrid[1:], columns=columns)
+        return self._get_table_and_convert(query, limit, index="mrid")
 
     def synchronous_machines(
         self,
-        sync_vars: Tuple[str] = ("sn", "p", "q"),
-        region: Union[str, List[str]] = None,
+        sync_vars: Iterable[SyncVars] = tuple(SyncVars),
+        region: Region = None,
         sub_region: bool = False,
         connectivity: Optional[str] = None,
+        nodes: Optional[str] = None,
         station_group_optional: bool = True,
         with_sequence_number: bool = False,
         network_analysis: bool = True,
         u_groups: bool = False,
-        mrid: str = mrid_variable,
-        name: str = "?name",
         limit: Optional[int] = None,
         dry_run: bool = False,
-    ) -> pd.DataFrame:
+        ssh_graph: Optional[str] = None,
+    ) -> Union[pd.DataFrame, str]:
         """Query synchronous machines
 
         Args:
@@ -229,38 +360,33 @@ class CimModel(Prefix):
            >>> gdbc = GraphDBClient(service('LATEST'))
            >>> gdbc.synchronous_machines(limit=10)
         """
-        terminal_mrid: str = "?t_mrid"
-
         query = queries.synchronous_machines_query(
             sync_vars,
             region,
             sub_region,
             connectivity,
+            nodes,
             station_group_optional,
             self.cim_version,
             with_sequence_number,
             network_analysis,
             u_groups,
-            terminal_mrid,
-            mrid,
-            name,
+            ssh_graph,
         )
         if dry_run:
             return self._query_with_header(query, limit)
-        float_list = ["maxP", "minP", "allocationMax", "allocationWeight", *sync_vars]
-        columns = {var: float for var in float_list}
-        return self._get_table_and_convert(query, limit, index=mrid[1:], columns=columns)
+        return self._get_table_and_convert(query, limit, index="mrid")
 
     def connections(
         self,
-        rdf_types: Union[str, Tuple[str]] = ("cim:Breaker", "cim:Disconnector"),
-        region: Optional[Union[str, List[str]]] = None,
+        rdf_types: Union[str, Iterable[str]] = ("cim:Breaker", "cim:Disconnector"),
+        region: Region = None,
         sub_region: bool = False,
         connectivity: Optional[str] = None,
-        mrid: str = mrid_variable,
+        nodes: Optional[str] = None,
         limit: Optional[int] = None,
         dry_run: bool = False,
-    ) -> pd.DataFrame:
+    ) -> Union[pd.DataFrame, str]:
         """Query connectors
 
         Args:
@@ -281,24 +407,23 @@ class CimModel(Prefix):
            >>> gdbc.connections(limit=10)
         """
         query = queries.connection_query(
-            self.cim_version, rdf_types, region, sub_region, connectivity, mrid
+            self.cim_version, rdf_types, region, sub_region, connectivity, nodes
         )
         if dry_run:
             return self._query_with_header(query, limit)
-        return self._get_table_and_convert(query, limit, index=mrid[1:])
+        return self._get_table_and_convert(query, limit, index="mrid")
 
     def borders(
         self,
-        region: Optional[Union[str, List[str]]] = None,
+        region: Union[str, List[str]],
         sub_region: bool = False,
+        nodes: Optional[str] = None,
         ignore_hvdc: bool = True,
         with_market_code: bool = False,
         market_optional: bool = False,
-        mrid: str = mrid_variable,
-        name: str = "?name",
         limit: Optional[int] = None,
         dry_run: bool = False,
-    ) -> pd.DataFrame:
+    ) -> Union[pd.DataFrame, str]:
         """Retrieve ACLineSegments where one terminal is inside and the other is outside the region
 
         Args:
@@ -311,77 +436,74 @@ class CimModel(Prefix):
             dry_run: return string with sql query
 
         """
-        query = queries.borders_query(
+        query = line_queries.borders_query(
             self.cim_version,
             region,
             sub_region,
+            nodes,
             ignore_hvdc,
             with_market_code,
             market_optional,
-            mrid,
-            name,
         )
         if dry_run:
             return self._query_with_header(query, limit)
-        return self._get_table_and_convert(query, limit, index=mrid[1:])
+        return self._get_table_and_convert(query, limit, index="mrid")
 
     def converters(
         self,
-        region: Optional[Union[str, List[str]]] = None,
+        region: Region = None,
         sub_region: bool = False,
-        converter_types: Iterable[str] = ("VoltageSource", "DC"),
-        mrid: str = mrid_variable,
-        name: str = "?name",
+        converter_types: Iterable[ConverterTypes] = tuple(ConverterTypes),
+        nodes: Optional[str] = None,
         sequence_numbers: Optional[List[int]] = None,
         dry_run: bool = False,
-    ) -> pd.DataFrame:
+    ) -> Union[pd.DataFrame, str]:
         query = queries.converters(
-            region, sub_region, converter_types, mrid, name, sequence_numbers
+            region, sub_region, self.in_prefixes(converter_types), nodes, sequence_numbers
         )
         if dry_run:
             return self._query_with_header(query)
-        return self._get_table_and_convert(query, index=mrid[1:])
+        return self._get_table_and_convert(query, index="mrid")
 
     def transformers_connected_to_converter(
         self,
-        region: Optional[Union[str, List[str]]] = None,
+        region: Region = None,
         sub_region: bool = False,
-        converter_types: Iterable[str] = ("VoltageSource", "DC"),
-        mrid: str = mrid_variable,
-        name: str = "?name",
+        converter_types: Iterable[ConverterTypes] = tuple(ConverterTypes),
         dry_run: bool = False,
-    ) -> pd.DataFrame:
+    ) -> Union[pd.DataFrame, str]:
         """Query list of transformer connected at a converter (Voltage source or DC)
 
         Args:
            region: Limit to region
            sub_region: Assume region is a sub_region
-           converter_types: VoltageSource or DC (ALG:VoltageSourceConverter or ALG:DCConverter)
+           converter_types: VoltageSource, DC (cim:VsConverter, cim:DCConverterUnit)
            dry_run: return string with sql query
 
         """
         query = queries.transformers_connected_to_converter(
-            region, sub_region, converter_types, mrid, name
+            region, sub_region, self.in_prefixes(converter_types)
         )
         if dry_run:
             return self._query_with_header(query)
-        return self._get_table_and_convert(query, index=mrid[1:])
+        return self._get_table_and_convert(query, index="mrid")
 
     def ac_lines(
         self,
-        region: Optional[Union[str, List[str]]] = None,
+        region: Region = None,
         sub_region: bool = False,
         limit: Optional[int] = None,
         connectivity: Optional[str] = None,
-        rates: Tuple[str] = ratings,
+        nodes: Optional[str] = None,
+        with_loss: bool = False,
+        rates: Iterable[Rates] = (Rates.Normal,),
         network_analysis: bool = True,
         with_market: bool = False,
         temperatures: Optional[List[int]] = None,
-        impedance: Iterable[str] = impedance_variables,
-        mrid: str = mrid_variable,
-        name: str = "?name",
+        impedance: Iterable[Impedance] = tuple(Impedance),
+        length: bool = False,
         dry_run: bool = False,
-    ) -> pd.DataFrame:
+    ) -> Union[pd.DataFrame, str]:
         """Query ac line segments
 
         Args:
@@ -400,19 +522,20 @@ class CimModel(Prefix):
            >>> gdbc = GraphDBClient(service('LATEST'))
            >>> gdbc.ac_lines(limit=10)
         """
-        query = queries.ac_line_query(
+        query = line_queries.ac_line_query(
             self.cim_version,
             self.ns["cim"],
             region,
             sub_region,
             connectivity,
+            nodes,
+            with_loss,
             rates,
             network_analysis,
             with_market,
             temperatures,
             impedance,
-            mrid,
-            name,
+            length,
         )
         if dry_run:
             return self._query_with_header(query, limit)
@@ -425,16 +548,18 @@ class CimModel(Prefix):
 
     def series_compensators(
         self,
-        region: Optional[Union[str, List[str]]] = None,
+        region: Region = None,
         sub_region: bool = False,
         connectivity: Optional[str] = None,
+        with_loss: bool = False,
+        nodes: Optional[str] = None,
+        rates: Iterable[Rates] = (Rates.Normal,),
         network_analysis: bool = True,
         with_market: bool = False,
-        mrid: str = mrid_variable,
-        name: str = "?name",
+        impedance: Iterable[Impedance] = tuple(Impedance),
         limit: Optional[int] = None,
         dry_run: bool = False,
-    ) -> pd.DataFrame:
+    ) -> Union[pd.DataFrame, str]:
         """Query series compensators
 
         Args:
@@ -452,15 +577,17 @@ class CimModel(Prefix):
            >>> gdbc = GraphDBClient(service('LATEST'))
            >>> gdbc.series_compensators(limit=10)
         """
-        query = queries.series_compensator_query(
+        query = line_queries.series_compensator_query(
             self.cim_version,
             region,
             sub_region,
             connectivity,
+            with_loss,
+            nodes,
+            rates,
             network_analysis,
             with_market,
-            mrid,
-            name,
+            impedance,
         )
         if dry_run:
             return self._query_with_header(query, limit)
@@ -468,18 +595,16 @@ class CimModel(Prefix):
 
     def transformers(
         self,
-        region: Optional[Union[str, List[str]]] = None,
+        region: Region = None,
         sub_region: bool = False,
         connectivity: Optional[str] = None,
-        rates: Tuple[str] = ratings,
+        rates: Iterable[Rates] = (Rates.Normal,),
         network_analysis: bool = True,
         with_market: bool = False,
-        impedance: Iterable[str] = impedance_variables,
-        mrid: str = "?p_mrid",
-        name: str = "?name",
+        impedance: Iterable[Impedance] = tuple(Impedance),
         limit: Optional[int] = None,
         dry_run: bool = False,
-    ) -> pd.DataFrame:
+    ) -> Union[pd.DataFrame, str]:
         """Query transformer windings
 
         Args:
@@ -500,15 +625,7 @@ class CimModel(Prefix):
            >>> gdbc.transformers(limit=10)
         """
         query = queries.transformer_query(
-            region,
-            sub_region,
-            connectivity,
-            rates,
-            network_analysis,
-            with_market,
-            mrid,
-            name,
-            impedance,
+            region, sub_region, connectivity, rates, network_analysis, with_market, impedance
         )
         if dry_run:
             return self._query_with_header(query, limit)
@@ -516,17 +633,19 @@ class CimModel(Prefix):
 
     def two_winding_transformers(
         self,
-        region: Optional[Union[str, List[str]]] = None,
+        region: Region = None,
         sub_region: bool = False,
-        rates: Tuple[str] = ratings,
+        rates: Iterable[Rates] = (Rates.Normal,),
         network_analysis: bool = True,
         with_market: bool = False,
-        impedance: Iterable[str] = impedance_variables,
-        mrid: str = "?p_mrid",
+        impedance: Iterable[Impedance] = tuple(Impedance),
+        p_mrid: bool = False,
+        nodes: Optional[str] = None,
+        with_loss: bool = False,
         name: str = "?name",
         limit: Optional[int] = None,
         dry_run: bool = False,
-    ) -> pd.DataFrame:
+    ) -> Union[pd.DataFrame, str]:
         """Query two-winding transformer
 
         Args:
@@ -546,7 +665,17 @@ class CimModel(Prefix):
            >>> gdbc.two_winding_transformers(limit=10)
         """
         query = queries.two_winding_transformer_query(
-            region, sub_region, rates, network_analysis, with_market, mrid, name, impedance
+            region,
+            sub_region,
+            rates,
+            network_analysis,
+            with_market,
+            p_mrid,
+            nodes,
+            with_loss,
+            name,
+            impedance,
+            self.cim_version,
         )
         if dry_run:
             return self._query_with_header(query, limit)
@@ -554,17 +683,19 @@ class CimModel(Prefix):
 
     def three_winding_transformers(
         self,
-        region: Optional[Union[str, List[str]]] = None,
+        region: Region = None,
         sub_region: bool = False,
-        rates: Tuple[str] = ratings,
+        rates: Iterable[Rates] = (Rates.Normal,),
         network_analysis: bool = True,
         with_market: bool = False,
-        impedance: Iterable[str] = impedance_variables,
-        mrid: str = "?p_mrid",
+        impedance: Iterable[Impedance] = tuple(Impedance),
+        p_mrid: bool = False,
+        nodes: Optional[str] = None,
+        with_loss: bool = False,
         name: str = "?name",
         limit: Optional[int] = None,
         dry_run: bool = False,
-    ) -> pd.DataFrame:
+    ) -> Union[pd.DataFrame, str]:
         """Query three-winding transformer. Return as three two-winding transformers.
 
         Args:
@@ -584,7 +715,17 @@ class CimModel(Prefix):
            >>> gdbc.two_winding_transformers(limit=10)
         """
         query = queries.three_winding_transformer_query(
-            region, sub_region, rates, network_analysis, with_market, mrid, name, impedance
+            region,
+            sub_region,
+            rates,
+            network_analysis,
+            with_market,
+            p_mrid,
+            nodes,
+            with_loss,
+            name,
+            impedance,
+            self.cim_version,
         )
         if dry_run:
             return self._query_with_header(query, limit)
@@ -592,7 +733,7 @@ class CimModel(Prefix):
 
     def disconnected(
         self, index: Optional[str] = None, limit: Optional[int] = None, dry_run: bool = False
-    ) -> pd.DataFrame:
+    ) -> Union[pd.DataFrame, str]:
         """Query disconneced status from ssh profile (not available in GraphDB)
 
         Args:
@@ -607,7 +748,7 @@ class CimModel(Prefix):
 
     def ssh_synchronous_machines(
         self, limit: Optional[int] = None, dry_run: bool = False
-    ) -> pd.DataFrame:
+    ) -> Union[pd.DataFrame, str]:
         """Query synchronous machines from ssh profile (not available in GraphDB)
 
         Args:
@@ -617,15 +758,14 @@ class CimModel(Prefix):
         query = ssh_queries.synchronous_machines()
         if dry_run:
             return self._query_with_header(query, limit)
-        columns = {"p": float, "q": float, "controlEnabled": bool}
-        return self._get_table_and_convert(query, limit, index=mrid_variable[1:], columns=columns)
+        return self._get_table_and_convert(query, limit, index="mrid")
 
     def ssh_load(
         self,
         rdf_types: Optional[List[str]] = None,
         limit: Optional[int] = None,
         dry_run: bool = False,
-    ) -> pd.DataFrame:
+    ) -> Union[pd.DataFrame, str]:
         """Query load data from ssh profile (not available in GraphDB)
 
         Args:
@@ -639,12 +779,14 @@ class CimModel(Prefix):
         query = ssh_queries.load(rdf_types)
         if dry_run:
             return self._query_with_header(query, limit)
-        columns = {"p": float, "q": float}
-        return self._get_table_and_convert(query, limit, index=mrid_variable[1:], columns=columns)
+        return self._get_table_and_convert(query, limit, index="mrid")
 
     def ssh_generating_unit(
-        self, rdf_types: List[str] = None, limit: Optional[int] = None, dry_run: bool = False
-    ) -> pd.DataFrame:
+        self,
+        rdf_types: Optional[List[str]] = None,
+        limit: Optional[int] = None,
+        dry_run: bool = False,
+    ) -> Union[pd.DataFrame, str]:
         """Query generating units from ssh profile (not available in GraphDB)
 
         Args:
@@ -655,14 +797,15 @@ class CimModel(Prefix):
 
         """
         if rdf_types is None:
-            rdf_types = [f"cim:{unit}GeneratingUnit" for unit in generating_types]
+            rdf_types = [f"cim:{unit}GeneratingUnit" for unit in GeneratorTypes]
         query = ssh_queries.generating_unit(rdf_types)
         if dry_run:
             return self._query_with_header(query, limit)
-        columns = {"normalPF": float}
-        return self._get_table_and_convert(query, limit, index=mrid_variable[1:], columns=columns)
+        return self._get_table_and_convert(query, limit, index="mrid")
 
-    def terminal(self, limit: Optional[int] = None, dry_run: bool = False) -> pd.DataFrame:
+    def terminal(
+        self, limit: Optional[int] = None, dry_run: bool = False
+    ) -> Union[pd.DataFrame, str]:
         """Query terminals from tp profile (not available in GraphDB)
 
         Args:
@@ -672,10 +815,11 @@ class CimModel(Prefix):
         query = tp_queries.terminal(self.cim_version)
         if dry_run:
             return self._query_with_header(query, limit)
-        columns = {"connected": bool}
-        return self._get_table_and_convert(query, limit, index=mrid_variable[1:], columns=columns)
+        return self._get_table_and_convert(query, limit, index="mrid")
 
-    def topological_node(self, limit: Optional[int] = None, dry_run: bool = False) -> pd.DataFrame:
+    def topological_node(
+        self, limit: Optional[int] = None, dry_run: bool = False
+    ) -> Union[pd.DataFrame, str]:
         """Query topological nodes from tp profile (not available in GraphDB)
 
         Args:
@@ -685,12 +829,14 @@ class CimModel(Prefix):
         query = tp_queries.topological_node()
         if dry_run:
             return self._query_with_header(query, limit)
-        columns = {"BaseVoltage": float}
-        return self._get_table_and_convert(query, limit, index=mrid_variable[1:], columns=columns)
+        return self._get_table_and_convert(query, limit, index="mrid")
 
     def powerflow(
-        self, power: Tuple[str] = ("p", "q"), limit: Optional[int] = None, dry_run: bool = False
-    ) -> pd.DataFrame:
+        self,
+        power: Tuple[str, ...] = ("p", "q"),
+        limit: Optional[int] = None,
+        dry_run: bool = False,
+    ) -> Union[pd.DataFrame, str]:
         """Query powerflow from sv profile (not available in GraphDB)
 
         Args:
@@ -701,15 +847,14 @@ class CimModel(Prefix):
         query = sv_queries.powerflow(power)
         if dry_run:
             return self._query_with_header(query, limit)
-        columns = {x: float for x in power}
-        return self._get_table_and_convert(query, limit, index=mrid_variable[1:], columns=columns)
+        return self._get_table_and_convert(query, limit, index="mrid")
 
     def voltage(
         self,
         voltage_vars: Iterable[str] = ("v", "angle"),
         limit: Optional[int] = None,
         dry_run: bool = False,
-    ) -> pd.DataFrame:
+    ) -> Union[pd.DataFrame, str]:
         """Query voltage from sv profile (not available in GraphDB)
 
         Args:
@@ -720,10 +865,11 @@ class CimModel(Prefix):
         query = sv_queries.voltage(voltage_vars)
         if dry_run:
             return self._query_with_header(query, limit)
-        columns = {x: float for x in voltage_vars}
-        return self._get_table_and_convert(query, limit, index=mrid_variable[1:], columns=columns)
+        return self._get_table_and_convert(query, limit, index="mrid")
 
-    def tapstep(self, limit: Optional[int] = None, dry_run: bool = False) -> pd.DataFrame:
+    def tapstep(
+        self, limit: Optional[int] = None, dry_run: bool = False
+    ) -> Union[pd.DataFrame, str]:
         """Query tapstep from sv profile (not available in GraphDB)
 
         Args:
@@ -733,8 +879,7 @@ class CimModel(Prefix):
         query = sv_queries.tapstep()
         if dry_run:
             return self._query_with_header(query, limit)
-        columns = {"position": float}
-        return self._get_table_and_convert(query, limit, index=mrid_variable[1:], columns=columns)
+        return self._get_table_and_convert(query, limit, index="mrid")
 
     @property
     def regions(self) -> pd.DataFrame:
@@ -749,115 +894,5 @@ class CimModel(Prefix):
            >>> gdbc = GraphDBClient(service('LATEST'))
            >>> gdbc.regions
         """
-        query = queries.regions_query(mrid_variable)
-        return self._get_table_and_convert(query, limit=None, index=mrid_variable[1:])
-
-    @property
-    def empty(self) -> bool:
-        """Identify empty GraphDB repo"""
-        try:
-            self._get_table("SELECT * \n WHERE { ?s ?p ?o }", limit=1)
-            return False
-        except IndexError:
-            return True
-
-    @staticmethod
-    def _assign_column_types(
-        result: pd.DataFrame, columns: Dict[str, Union[bool, str, float, int]]
-    ) -> None:
-        for column, column_type in columns.items():
-            if column_type is str:
-                continue
-            if column_type is bool:
-                result[column] = result[column].str.contains("true", flags=re.IGNORECASE)
-            else:
-                result.loc[result[column] == "None", column] = ""
-                result[column] = pd.to_numeric(result[column], errors="coerce").astype(column_type)
-
-    @classmethod
-    def col_map(cls, data_row, columns) -> Tuple[Dict[str, str]]:
-        columns = {} if columns is None else columns
-        col_map = cls._col_map(data_row)
-        return col_map, {col: columns[col] for col in set(columns).difference(col_map)}
-
-    def get_table(
-        self,
-        query: str,
-        index: Optional[str] = None,
-        limit: Optional[int] = None,
-        map_data_types: bool = True,
-        custom_maps: Optional[Dict] = None,
-        columns: Optional[Dict] = None,
-    ) -> pd.DataFrame:
-        """Gets given table from the configured database.
-
-        Args:
-           query: to sparql server
-           index: column name to use as index
-           limit: limit number of resulting rows
-           map_data_types: gets datatypes from the configured graphdb & maps the types in the result
-                  to correct python types
-           custom_maps: dictionary of 'sparql_datatype': function to apply on columns with that
-                  type. Overwrites sparql map for the types specified.
-           columns: dictionary of 'column_name': function,
-                  uses pandas astype on the column, or applies function.
-                  Sparql map overwrites columns when available
-
-        Example:
-           >>> from cimsparql.graphdb import GraphDBClient
-           >>> from cimsparql.url import service
-           >>> gdbc = GraphDBClient(service('LATEST'))
-           >>> query = 'select * where { ?subject ?predicate ?object }'
-           >>> gdbc.get_table(query, limit=10)
-        """
-        try:
-            result, data_row = self._get_table(query, limit)
-        except IndexError:
-            return pd.DataFrame([])
-
-        if map_data_types and self.mapper is not None:
-            col_map, columns = self.col_map(data_row, columns)
-            result = self.mapper.map_data_types(result, col_map, custom_maps, columns)
-
-        if index and not result.empty:
-            result.set_index(index, inplace=True)
-        return result
-
-    @property
-    def map_data_types(self) -> bool:
-        try:
-            return self.mapper.have_cim_version(self.prefixes["cim"])
-        except AttributeError:
-            return False
-
-    @classmethod
-    def _manual_convert_types(
-        cls: CimModelType, df: pd.DataFrame, columns: Dict, index: str
-    ) -> pd.DataFrame:
-        if columns is None:
-            columns = {}
-        reset_index = index is not None
-        if reset_index:
-            df.reset_index(inplace=True)
-        df = df.astype(str)
-        cls._assign_column_types(df, columns)
-        if reset_index:
-            df.set_index(index, inplace=True)
-        return df
-
-    def _get_table_and_convert(
-        self,
-        query: str,
-        limit: Optional[int] = None,
-        index: Optional[str] = None,
-        custom_maps: Optional[Dict] = None,
-        columns: Optional[Dict] = None,
-    ) -> pd.DataFrame:
-        result = self.get_table(
-            query, index, limit, map_data_types=True, custom_maps=custom_maps, columns=columns
-        )
-
-        if not self.map_data_types and len(result) > 0:
-            result = self._manual_convert_types(result, columns, index)
-
-        return result
+        query = queries.regions_query()
+        return self._get_table_and_convert(query, limit=None, index="mrid")
